@@ -27,20 +27,32 @@
 
 #include "glib.h"
 #include "gunichartables.h"
+#include "gunicodeprivate.h"
 
+#define ATTR_TABLE(Page) (((Page) <= G_UNICODE_LAST_PAGE_PART1) \
+                          ? attr_table_part1[Page] \
+                          : attr_table_part2[(Page) - 0xe00])
 
 #define ATTTABLE(Page, Char) \
-  ((attr_table[Page] == G_UNICODE_MAX_TABLE_INDEX) ? 0 : (attr_data[attr_table[Page]][Char]))
+  ((ATTR_TABLE(Page) == G_UNICODE_MAX_TABLE_INDEX) ? 0 : (attr_data[ATTR_TABLE(Page)][Char]))
 
-/* We cheat a bit and cast type values to (char *).  We detect these
-   using the &0xff trick.  */
-#define TTYPE(Page, Char) \
-  ((type_table[Page] >= G_UNICODE_MAX_TABLE_INDEX) \
-   ? (type_table[Page] - G_UNICODE_MAX_TABLE_INDEX) \
-   : (type_data[type_table[Page]][Char]))
+#define TTYPE_PART1(Page, Char) \
+  ((type_table_part1[Page] >= G_UNICODE_MAX_TABLE_INDEX) \
+   ? (type_table_part1[Page] - G_UNICODE_MAX_TABLE_INDEX) \
+   : (type_data[type_table_part1[Page]][Char]))
 
+#define TTYPE_PART2(Page, Char) \
+  ((type_table_part2[Page] >= G_UNICODE_MAX_TABLE_INDEX) \
+   ? (type_table_part2[Page] - G_UNICODE_MAX_TABLE_INDEX) \
+   : (type_data[type_table_part2[Page]][Char]))
+
+#define TYPE(Char) \
+  (((Char) <= G_UNICODE_LAST_CHAR_PART1) \
+   ? TTYPE_PART1 ((Char) >> 8, (Char) & 0xff) \
+   : (((Char) >= 0xe0000 && (Char) <= G_UNICODE_LAST_CHAR) \
+      ? TTYPE_PART2 (((Char) - 0xe0000) >> 8, (Char) & 0xff) \
+      : G_UNICODE_UNASSIGNED))
 
-#define TYPE(Char) (((Char) > (G_UNICODE_LAST_CHAR)) ? G_UNICODE_UNASSIGNED : TTYPE ((Char) >> 8, (Char) & 0xff))
 
 #define ISDIGIT(Type) ((Type) == G_UNICODE_DECIMAL_NUMBER	\
 		       || (Type) == G_UNICODE_LETTER_NUMBER	\
@@ -363,10 +375,10 @@ g_unichar_toupper (gunichar c)
   if (t == G_UNICODE_LOWERCASE_LETTER)
     {
       gunichar val = ATTTABLE (c >> 8, c & 0xff);
-      if (val >= 0xd800 && val < 0xdc00)
+      if (val >= 0x1000000)
 	{
-	  const guchar *p = special_case_table[val - 0xd800];
-	  return p[0] * 256 + p[1];
+	  const guchar *p = special_case_table + val - 0x1000000;
+	  return g_utf8_get_char (p);
 	}
       else
 	return val ? val : c;
@@ -400,10 +412,10 @@ g_unichar_tolower (gunichar c)
   if (t == G_UNICODE_UPPERCASE_LETTER)
     {
       gunichar val = ATTTABLE (c >> 8, c & 0xff);
-      if (val >= 0xd800 && val < 0xdc00)
+      if (val >= 0x1000000)
 	{
-	  const guchar *p = special_case_table[val - 0xd800];
-	  return p[0] * 256 + p[1];
+	  const guchar *p = special_case_table + val - 0x1000000;
+	  return g_utf8_get_char (p);
 	}
       else
 	return val ? val : c;
@@ -533,7 +545,7 @@ get_locale_type (void)
   return LOCALE_NORMAL;
 }
 
-static int
+static gint
 output_marks (const char **p_inout,
 	      char        *out_buffer,
 	      gboolean     remove_dot)
@@ -562,32 +574,22 @@ output_marks (const char **p_inout,
 
 static gint
 output_special_case (gchar *out_buffer,
-		     int    index,
+		     int    offset,
 		     int    type,
 		     int    which)
 {
-  const guchar *p = special_case_table[index];
-  gint len = 0;
+  const guchar *p = special_case_table + offset;
+  gint len;
 
   if (type != G_UNICODE_TITLECASE_LETTER)
-    p += 2; /* +2 to skip over "best single match" */
+    p = g_utf8_next_char (p);
 
   if (which == 1)
-    {
-      while (p[0] || p[1])
-	p += 2;
-      p += 2;
-    }
+    p += strlen (p) + 1;
 
-  while (TRUE)
-    {
-      gunichar ch = p[0] * 256 + p[1];
-      if (!ch)
-	break;
-
-      len += g_unichar_to_utf8 (ch, out_buffer ? out_buffer + len : NULL);
-      p += 2;
-    }
+  len = strlen (p);
+  if (out_buffer)
+    memcpy (out_buffer, p, len);
 
   return len;
 }
@@ -664,9 +666,9 @@ real_toupper (const gchar *str,
 	{
 	  val = ATTTABLE (c >> 8, c & 0xff);
 
-	  if (val >= 0xd800 && val < 0xdc00)
+	  if (val >= 0x1000000)
 	    {
-	      len += output_special_case (out_buffer ? out_buffer + len : NULL, val - 0xd800, t,
+	      len += output_special_case (out_buffer ? out_buffer + len : NULL, val - 0x1000000, t,
 					  t == G_UNICODE_LOWERCASE_LETTER ? 0 : 1);
 	    }
 	  else
@@ -736,6 +738,28 @@ g_utf8_strup (const gchar *str,
   return result;
 }
 
+/* traverses the string checking for characters with combining class == 230
+ * until a base character is found */
+static gboolean
+has_more_above (const gchar *str)
+{
+  const gchar *p = str;
+  gint combining_class;
+
+  while (*p)
+    {
+      combining_class = _g_unichar_combining_class (g_utf8_get_char (p));
+      if (combining_class == 230)
+        return TRUE;
+      else if (combining_class == 0)
+        break;
+
+      p = g_utf8_next_char (p);
+    }
+
+  return FALSE;
+}
+
 static gsize
 real_tolower (const gchar *str,
 	      gssize       max_len,
@@ -757,9 +781,46 @@ real_tolower (const gchar *str,
 
       if (locale_type == LOCALE_TURKIC && c == 'I')
 	{
-	  /* I => LATIN SMALL LETTER DOTLESS I */
-	  len += g_unichar_to_utf8 (0x131, out_buffer ? out_buffer + len : NULL); 
-	}
+          if (g_utf8_get_char (p) == 0x0307)
+            {
+              /* I + COMBINING DOT ABOVE => i (U+0069) */
+              len += g_unichar_to_utf8 (0x0069, out_buffer ? out_buffer + len : NULL); 
+              p = g_utf8_next_char (p);
+            }
+          else
+            {
+              /* I => LATIN SMALL LETTER DOTLESS I */
+              len += g_unichar_to_utf8 (0x131, out_buffer ? out_buffer + len : NULL); 
+            }
+        }
+      /* Introduce an explicit dot above when lowercasing capital I's and J's
+       * whenever there are more accents above. [SpecialCasing.txt] */
+      else if (locale_type == LOCALE_LITHUANIAN && 
+               (c == 0x00cc || c == 0x00cd || c == 0x0128))
+        {
+          len += g_unichar_to_utf8 (0x0069, out_buffer ? out_buffer + len : NULL); 
+          len += g_unichar_to_utf8 (0x0307, out_buffer ? out_buffer + len : NULL); 
+
+          switch (c)
+            {
+            case 0x00cc: 
+              len += g_unichar_to_utf8 (0x0300, out_buffer ? out_buffer + len : NULL); 
+              break;
+            case 0x00cd: 
+              len += g_unichar_to_utf8 (0x0301, out_buffer ? out_buffer + len : NULL); 
+              break;
+            case 0x0128: 
+              len += g_unichar_to_utf8 (0x0303, out_buffer ? out_buffer + len : NULL); 
+              break;
+            }
+        }
+      else if (locale_type == LOCALE_LITHUANIAN && 
+               (c == 'I' || c == 'J' || c == 0x012e) && 
+               has_more_above (p))
+        {
+          len += g_unichar_to_utf8 (g_unichar_tolower (c), out_buffer ? out_buffer + len : NULL); 
+          len += g_unichar_to_utf8 (0x0307, out_buffer ? out_buffer + len : NULL); 
+        }
       else if (c == 0x03A3)	/* GREEK CAPITAL LETTER SIGMA */
 	{
 	  if ((max_len < 0 || p < str + max_len) && *p)
@@ -787,9 +848,9 @@ real_tolower (const gchar *str,
 	{
 	  val = ATTTABLE (c >> 8, c & 0xff);
 
-	  if (val >= 0xd800 && val < 0xdc00)
+	  if (val >= 0x1000000)
 	    {
-	      len += output_special_case (out_buffer ? out_buffer + len : NULL, val - 0xd800, t, 0);
+	      len += output_special_case (out_buffer ? out_buffer + len : NULL, val - 0x1000000, t, 0);
 	    }
 	  else
 	    {
@@ -884,6 +945,8 @@ g_utf8_casefold (const gchar *str,
   GString *result = g_string_new (NULL);
   const char *p;
 
+  g_return_val_if_fail (str != NULL, NULL);
+
   p = str;
   while ((len < 0 || p < str + len) && *p)
     {
@@ -893,7 +956,7 @@ g_utf8_casefold (const gchar *str,
       int end = G_N_ELEMENTS (casefold_table);
 
       if (ch >= casefold_table[start].ch &&
-	  ch <= casefold_table[end - 1].ch)
+          ch <= casefold_table[end - 1].ch)
 	{
 	  while (TRUE)
 	    {
@@ -920,3 +983,64 @@ g_utf8_casefold (const gchar *str,
 
   return g_string_free (result, FALSE); 
 }
+
+/**
+ * g_unichar_get_mirror_char:
+ * @ch: a unicode character
+ * @mirrored_ch: location to store the mirrored character
+ * 
+ * In Unicode, some characters are <firstterm>mirrored</firstterm>. This
+ * means that their images are mirrored horizontally in text that is laid
+ * out from right to left. For instance, "(" would become its mirror image,
+ * ")", in right-to-left text.
+ *
+ * If @ch has the Unicode mirrored property and there is another unicode
+ * character that typically has a glyph that is the mirror image of @ch's
+ * glyph, puts that character in the address pointed to by @mirrored_ch.
+ *
+ * Return value: %TRUE if @ch has a mirrored character and @mirrored_ch is
+ * filled in, %FALSE otherwise
+ *
+ * Since: 2.4
+ **/
+/* This code is adapted from FriBidi (http://fribidi.sourceforge.net/). 
+ * FriBidi is: Copyright (C) 1999,2000 Dov Grobgeld, and
+ *             Copyright (C) 2001,2002 Behdad Esfahbod.
+ */
+gboolean
+g_unichar_get_mirror_char (gunichar ch,
+                           gunichar *mirrored_ch)
+{
+  gint pos, step, size;
+  gboolean found;
+
+  size = G_N_ELEMENTS (bidi_mirroring_table);
+  pos = step = (size / 2) + 1;
+
+  while (step > 1)
+    {
+      gunichar cmp_ch = bidi_mirroring_table[pos].ch;
+      step = (step + 1) / 2;
+
+      if (cmp_ch < ch)
+        {
+          pos += step;
+          if (pos > size - 1)
+            pos = size - 1;
+        }
+      else if (cmp_ch > ch)
+        {
+          pos -= step;
+          if (pos < 0)
+            pos = 0;
+        }
+      else
+        break;
+    }
+  found = bidi_mirroring_table[pos].ch == ch;
+  if (mirrored_ch)
+    *mirrored_ch = found ? bidi_mirroring_table[pos].mirrored_ch : ch;
+
+  return found;
+
+}
